from __future__ import annotations

from loguru import logger

from avilla.core.account import AccountInfo
from avilla.core.platform import Abstract, Land, Platform, Version
from avilla.core.selector import Selector
from avilla.red.account import RedAccount
from avilla.red.capability import RedCapability
from avilla.red.collector.connection import ConnectionCollector
from avilla.standard.core.account.event import AccountAvailable, AccountRegistered


class RedEventLifespanPerform((m := ConnectionCollector())._):
    m.namespace = "avilla.protocol/red::event"
    m.identify = "lifespan"

    @m.entity(RedCapability.event_callback, event_type="meta::connect")
    async def connect(self, event_type: ..., raw_event: dict):
        self_id: int = raw_event["authData"]["account"]
        account = self.connection.account

        if account is None:
            # create account instance
            account = RedAccount(route=Selector().land("qq").account(str(self_id)), protocol=self.protocol)
            version_info = raw_event["version"]
            platform = Platform(
                Land("qq"),
                Abstract(f"red-protocol/{version_info}"),
                Version({"app": version_info}),
            )
            self.connection.account = account
            self.protocol.avilla.accounts[account.route] = AccountInfo(account.route, account, self.protocol, platform)
            account.websocket_client = self.connection
            logger.success(f"Account {self_id} connected and created with red-protocol version {version_info}")
            return AccountRegistered(self.protocol.avilla, account)
        else:
            self.connection.account = account
            account.route = Selector().land("qq").account(str(self_id))
            return AccountAvailable(self.protocol.avilla, account)



"""Ariadne 的异常定义"""


from avilla.core.exceptions import (
    InvalidAuthentication,
    InvalidOperation,
    ParserException,
)


class InvalidEventTypeDefinition(ParserException):
    """不合法的事件类型定义."""


class InvalidVerifyKey(InvalidAuthentication):
    """无效的 verifyKey 或其配置."""


class AccountNotFound(InvalidAuthentication):
    """未能使用所配置的账号激活 sessionKey, 请检查 mirai_session 配置."""


class InvalidSession(InvalidAuthentication):
    """无效的 sessionKey, 请重新获取."""


class UnVerifiedSession(InvalidAuthentication):
    """尚未验证/绑定的 session."""


class MissingNecessaryArgument(InvalidOperation):
    """应在所提到的参数之中至少传入/使用一个"""


class ConflictItem(InvalidOperation):
    """项冲突/其中一项被重复定义"""


def permission_error_message(operator: str, current: str, required: list[str]):
    return (
        f"missing permission: {operator} required "
        + " or ".join(f'"{i}"' for i in required)
        + f" but current is {current}"
    )



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from graia.amnesia.builtins.memcache import Memcache, MemcacheService

from avilla.core.exceptions import UnknownTarget
from avilla.core.ryanvk.collector.account import AccountCollector
from avilla.core.selector import Selector
from avilla.standard.core.profile import Nick, Summary

if TYPE_CHECKING:
    from avilla.red.account import RedAccount  # noqa
    from avilla.red.protocol import RedProtocol  # noqa


class RedFriendActionPerform((m := AccountCollector["RedProtocol", "RedAccount"]())._):
    m.namespace = "avilla.protocol/red::action"
    m.identify = "friend"

    @m.pull("land.friend", Summary)
    async def get_summary(self, target: Selector, route: ...) -> Summary:
        cache: Memcache = self.protocol.avilla.launch_manager.get_component(MemcacheService).cache
        if raw := await cache.get(f"red/account({self.account.route['account']}).friend({target.pattern['friend']})"):
            return Summary(raw["nick"], "a friend contact assigned to this account")
        result = await self.account.websocket_client.call_http("get", "api/bot/friends", {})
        result = cast(list, result)
        for i in result:
            friend_id = str(i["uin"])
            if friend_id == target["friend"]:
                return Summary(i["nick"], "name of friend")
        raise UnknownTarget("Friend not found")

    @m.pull("land.friend", Nick)
    async def get_nick(self, target: Selector, route: ...) -> Nick:
        cache: Memcache = self.protocol.avilla.launch_manager.get_component(MemcacheService).cache
        if raw := await cache.get(f"red/account({self.account.route['account']}).friend({target.pattern['friend']})"):
            return Nick(raw["nick"], raw["remark"] or raw["nick"], raw["longNick"])
        result = await self.account.websocket_client.call_http("get", "api/bot/friends", {})
        result = cast(list, result)
        for i in result:
            friend_id = str(i["uin"])
            if friend_id == target["friend"]:
                return Nick(i["nick"], i["remark"], i["longNick"])
        raise UnknownTarget("Friend not found")



from __future__ import annotations

import asyncio
import json
from typing import TYPE_CHECKING, Literal, cast

import aiohttp
from launart import Service
from launart.manager import Launart
from launart.utilles import any_completed
from loguru import logger

from avilla.red.account import RedAccount
from avilla.red.net.base import RedNetworking
from avilla.standard.core.account import AccountUnavailable, AccountUnregistered

if TYPE_CHECKING:
    from avilla.red.protocol import RedConfig, RedProtocol


class RedWsClientNetworking(RedNetworking, Service):
    required: set[str] = set()
    stages: set[str] = {"preparing", "blocking", "cleanup"}

    config: RedConfig
    connection: aiohttp.ClientWebSocketResponse | None = None
    session: aiohttp.ClientSession

    def __init__(self, protocol: RedProtocol, config: RedConfig) -> None:
        super().__init__(protocol)
        self.config = config

    @property
    def id(self):
        return f"satori/connection/websocket/client#{id(self)}"

    async def message_receive(self):
        if self.connection is None:
            raise RuntimeError("connection is not established")

        async for msg in self.connection:
            if msg.type in {aiohttp.WSMsgType.CLOSE, aiohttp.WSMsgType.ERROR, aiohttp.WSMsgType.CLOSED}:
                self.close_signal.set()
                break
            elif msg.type == aiohttp.WSMsgType.TEXT:
                data: dict = json.loads(cast(str, msg.data))
                yield self, data
        else:
            await self.connection_closed()

    async def send(self, payload: dict):
        if self.connection is None:
            raise RuntimeError("connection is not established")

        await self.connection.send_json(payload)

    async def call_http(
        self, method: Literal["get", "post", "multipart"], action: str, params: dict | None = None, raw: bool = False
    ):
        action = action.replace("_", "/")
        if method == "get":
            async with self.session.get(
                (self.config.http_endpoint / action).with_query(params or {}),
                headers={"Authorization": f"Bearer {self.config.access_token}"},
            ) as resp:
                return (await resp.content.read()) if raw else await resp.json(content_type=None)
        if method == "post":
            async with self.session.post(
                (self.config.http_endpoint / action),
                json=params or {},
                headers={"Authorization": f"Bearer {self.config.access_token}"},
            ) as resp:
                return (await resp.content.read()) if raw else await resp.json(content_type=None)
        if method == "multipart":
            data = aiohttp.FormData(quote_fields=False)
            if params is None:
                raise TypeError("multipart requires params")
            for k, v in params.items():
                if isinstance(v, dict):
                    data.add_field(k, v["value"], filename=v.get("filename"), content_type=v.get("content_type"))
                else:
                    data.add_field(k, v)

            async with self.session.post(
                (self.config.http_endpoint / action),
                data=data,
                headers={"Authorization": f"Bearer {self.config.access_token}"},
            ) as resp:
                return (await resp.content.read()) if raw else await resp.json(content_type=None)
        raise ValueError(f"Unknown method {method}")

    async def wait_for_available(self):
        await self.status.wait_for_available()

    def get_staff_components(self):
        return {"connection": self, "protocol": self.protocol, "avilla": self.protocol.avilla}

    def __staff_generic__(self, element_type: dict, event_type: dict):
        ...

    def get_staff_artifacts(self):
        return [self.protocol.artifacts, self.protocol.avilla.global_artifacts]

    @property
    def alive(self):
        return self.connection is not None and not self.connection.closed

    async def connection_daemon(self, manager: Launart, session: aiohttp.ClientSession):
        while not manager.status.exiting:
            try:
                async with session.ws_connect(self.config.endpoint, timeout=30) as self.connection:
                    logger.info(f"{self} Websocket client connected")
                    self.close_signal.clear()

                    close_task = asyncio.create_task(self.close_signal.wait())
                    receiver_task = asyncio.create_task(self.message_handle())
                    sigexit_task = asyncio.create_task(manager.status.wait_for_sigexit())
                    await self.send(
                        {
                            "type": "meta::connect",
                            "payload": {
                                "token": self.config.access_token,
                            },
                        }
                    )
                    done, pending = await any_completed(
                        sigexit_task,
                        close_task,
                        receiver_task,
                    )
                    avilla = self.protocol.avilla
                    if sigexit_task in done:
                        logger.info(f"{self} Websocket client exiting...")
                        await self.connection.close()
                        self.close_signal.set()
                        self.connection = None
                        for v in list(avilla.accounts.values()):
                            if v.protocol is self.protocol and self.account and v.route["account"] == self.account.route["account"]:  # type: ignore
                                self.account = None
                                del avilla.accounts[v.route]
                                await avilla.broadcast.postEvent(AccountUnregistered(avilla, v.account))
                                return
                    if close_task in done:
                        receiver_task.cancel()
                        logger.warning(f"{self} Connection closed by server, will reconnect in 5 seconds...")
                        for n in list(avilla.accounts.keys()):
                            logger.debug(f"Unregistering red-protocol account {n}...")
                            account = cast("RedAccount", avilla.accounts[n].account)
                            account.status.enabled = False
                            await avilla.broadcast.postEvent(AccountUnavailable(avilla, avilla.accounts[n].account))
                        self.account = None
                        await asyncio.sleep(5)
                        logger.info(f"{self} Reconnecting...")
                        continue
            except Exception as e:
                logger.error(f"{self} Error while connecting: {e}")
                await asyncio.sleep(5)
                logger.info(f"{self} Reconnecting...")

    async def launch(self, manager: Launart):
        async with self.stage("preparing"):
            self.session = aiohttp.ClientSession()

        async with self.stage("blocking"):
            await self.connection_daemon(manager, self.session)

        async with self.stage("cleanup"):
            await self.session.close()
            self.connection = None



from __future__ import annotations

from datetime import timedelta

from loguru import logger

from avilla.core.context import Context
from avilla.core.event import MetadataModified, ModifyDetail
from avilla.core.selector import Selector
from avilla.red.capability import RedCapability
from avilla.red.collector.connection import ConnectionCollector
from avilla.standard.core.privilege import MuteInfo


class RedEventGroupMemberPerform((m := ConnectionCollector())._):
    m.namespace = "avilla.protocol/red::event"
    m.identify = "member"

    @m.entity(RedCapability.event_callback, event_type="group::member::mute")
    async def group_name_change(self, event_type: ..., raw_event: dict):
        account = self.connection.account
        if account is None:
            logger.warning(f"Unknown account received message {raw_event}")
            return
        group = Selector().land(account.route["land"]).group(str(raw_event.get("peerUin", raw_event.get("peerUid"))))
        group_data = raw_event["elements"][0]["grayTipElement"]["groupElement"]
        target = group.member(str(group_data["shutUp"]["member"]["uin"]))
        operator = group.member(str(group_data["shutUp"]["admin"]["uin"]))
        context = Context(
            account,
            operator,
            target,
            group,
            group.member(account.route["account"]),
        )
        if group_data["shutUp"]["duration"] == "0":
            return MetadataModified(
                context,
                target,
                MuteInfo,
                {
                    MuteInfo.inh().muted: ModifyDetail("update", False, True),
                    MuteInfo.inh().duration: ModifyDetail("clear", timedelta(seconds=0)),
                },
                operator=operator,
                scene=group,
            )
        else:
            return MetadataModified(
                context,
                target,
                MuteInfo,
                {
                    MuteInfo.inh().muted: ModifyDetail("update", True, False),
                    MuteInfo.inh().duration: ModifyDetail(
                        "set", timedelta(seconds=int(group_data["shutUp"]["duration"]))
                    ),
                },
                operator=operator,
                scene=group,
            )


_ = {
    "msgId": "7273392328887159418",
    "msgRandom": "871925483165536739",
    "msgSeq": "6285",
    "cntSeq": "0",
    "chatType": 2,
    "msgType": 5,
    "subMsgType": 8,
    "sendType": 3,
    "senderUid": "",
    "peerUid": "592387986",
    "channelId": "",
    "guildId": "",
    "guildCode": "0",
    "fromUid": "0",
    "fromAppid": "0",
    "msgTime": "1693468617",
    "msgMeta": "0x",
    "sendStatus": 2,
    "sendMemberName": "",
    "sendNickName": "",
    "guildName": "",
    "channelName": "",
    "elements": [
        {
            "elementType": 8,
            "elementId": "7273392328887159419",
            "extBufForUI": "0x",
            "textElement": None,
            "faceElement": None,
            "marketFaceElement": None,
            "replyElement": None,
            "picElement": None,
            "pttElement": None,
            "videoElement": None,
            "grayTipElement": {
                "subElementType": 4,
                "revokeElement": None,
                "proclamationElement": None,
                "emojiReplyElement": None,
                "groupElement": {
                    "type": 8,
                    "role": 0,
                    "groupName": "",
                    "memberUid": "",
                    "memberNick": "",
                    "memberRemark": "",
                    "adminUid": "",
                    "adminNick": "",
                    "adminRemark": "",
                    "createGroup": None,
                    "memberAdd": None,
                    "shutUp": {
                        "curTime": "1693468617",
                        "duration": "600",
                        "admin": {
                            "uid": "u_ZLOOr1URIwoiy98-sw6eew",
                            "card": "",
                            "name": "",
                            "role": 4,
                            "uin": "3165388245",
                        },
                        "member": {
                            "uid": "u_pj2er5ElsEqyyImT-8ogeA",
                            "card": "",
                            "name": "小号机3",
                            "role": 0,
                            "uin": "3542335634",
                        },
                    },
                    "memberUin": "516689493",
                    "adminUin": "516689493",
                },
                "buddyElement": None,
                "feedMsgElement": None,
                "essenceElement": None,
                "groupNotifyElement": None,
                "buddyNotifyElement": None,
                "xmlElement": None,
                "fileReceiptElement": None,
                "localGrayTipElement": None,
                "blockGrayTipElement": None,
                "aioOpGrayTipElement": None,
                "jsonGrayTipElement": None,
            },
            "arkElement": None,
            "fileElement": None,
            "liveGiftElement": None,
            "markdownElement": None,
            "structLongMsgElement": None,
            "multiForwardMsgElement": None,
            "giphyElement": None,
            "walletElement": None,
            "inlineKeyboardElement": None,
            "textGiftElement": None,
            "calendarElement": None,
            "yoloGameResultElement": None,
            "avRecordElement": None,
        }
    ],
    "records": [],
    "emojiLikesList": [],
    "commentCnt": "0",
    "directMsgFlag": 0,
    "directMsgMembers": [],
    "peerName": "",
    "freqLimitInfo": None,
    "editable": False,
    "avatarMeta": "",
    "avatarPendant": "",
    "feedId": "",
    "roleId": "0",
    "timeStamp": "0",
    "clientIdentityInfo": None,
    "isImportMsg": False,
    "atType": 0,
    "fromChannelRoleInfo": {"roleId": "0", "name": "", "color": 0},
    "fromGuildRoleInfo": {"roleId": "0", "name": "", "color": 0},
    "levelRoleInfo": {"roleId": "0", "name": "", "color": 0},
    "recallTime": "0",
    "isOnlineMsg": False,
    "generalFlags": "0x",
    "clientSeq": "0",
    "fileGroupSize": None,
    "foldingInfo": None,
    "nameType": 0,
    "avatarFlag": 0,
    "anonymousExtInfo": None,
    "personalMedal": None,
    "roleManagementTag": None,
    "senderUin": "516689493",
    "peerUin": "592387986",
}



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from graia.amnesia.builtins.memcache import Memcache, MemcacheService

from avilla.core.exceptions import UnknownTarget
from avilla.core.ryanvk.collector.account import AccountCollector
from avilla.core.selector import Selector
from avilla.standard.core.common import Count
from avilla.standard.core.privilege import MuteAllCapability
from avilla.standard.core.profile import Nick, Summary
from avilla.standard.core.relation import SceneCapability

if TYPE_CHECKING:
    from avilla.red.account import RedAccount  # noqa
    from avilla.red.protocol import RedProtocol  # noqa


class RedGroupActionPerform((m := AccountCollector["RedProtocol", "RedAccount"]())._):
    m.namespace = "avilla.protocol/red::action"
    m.identify = "group"

    @m.pull("land.group", Summary)
    async def get_summary(self, target: Selector, route: ...) -> Summary:
        cache: Memcache = self.protocol.avilla.launch_manager.get_component(MemcacheService).cache
        if raw := await cache.get(f"red/account({self.account.route['account']}).group({target.pattern['group']})"):
            return Summary(raw["name"], "a group contact assigned to this account")
        result = await self.account.websocket_client.call_http("get", "api/bot/groups", {})
        result = cast(list, result)
        for i in result:
            group_id = str(i["groupCode"])
            if group_id == target["group"]:
                return Summary(i["groupName"], "name of group")
        raise UnknownTarget("Group not found")

    @m.pull("land.group", Nick)
    async def get_nick(self, target: Selector, route: ...) -> Nick:
        cache: Memcache = self.protocol.avilla.launch_manager.get_component(MemcacheService).cache
        if raw := await cache.get(f"red/account({self.account.route['account']}).group({target.pattern['group']})"):
            return Nick(raw["name"], raw["remark"] or raw["name"], None)
        result = await self.account.websocket_client.call_http("get", "api/bot/groups", {})
        result = cast(list, result)
        for i in result:
            group_id = str(i["groupCode"])
            if group_id == target["group"]:
                return Nick(i["groupName"], i["remarkName"], None)
        raise UnknownTarget("Group not found")

    @m.pull("land.group", Count)
    async def get_count(self, target: Selector, route: ...) -> Count:
        cache: Memcache = self.protocol.avilla.launch_manager.get_component(MemcacheService).cache
        if raw := await cache.get(f"red/account({self.account.route['account']}).group({target.pattern['group']})"):
            return Count(raw["memberCount"], raw["maxMember"])
        result = await self.account.websocket_client.call_http("get", "api/bot/groups", {})
        result = cast(list, result)
        for i in result:
            group_id = str(i["groupCode"])
            if group_id == target["group"]:
                return Count(i["memberCount"], i["maxMember"])
        raise UnknownTarget("Group not found")

    @m.entity(MuteAllCapability.mute_all, target="land.group")
    async def mute_all(self, target: Selector):
        await self.account.websocket_client.call_http(
            "post", "api/group/muteEveryone", {"group": int(target["group"]), "enable": True}
        )

    @m.entity(MuteAllCapability.unmute_all, target="land.group")
    async def unmute_all(self, target: Selector):
        await self.account.websocket_client.call_http(
            "post", "api/group/muteEveryone", {"group": int(target["group"]), "enable": False}
        )

    @m.entity(SceneCapability.remove_member, target="land.group.member")
    async def remove_member(self, target: Selector, reason: str | None = None):
        await self.account.websocket_client.call_http(
            "post",
            "api/group/kick",
            {
                "group": int(target["group"]),
                "uidList": [int(target["member"])],
                "reason": reason or "",
                "refuseForever": False,
            },
        )



from __future__ import annotations

from datetime import timedelta
from typing import TYPE_CHECKING, cast

from graia.amnesia.builtins.memcache import Memcache, MemcacheService

from avilla.core.exceptions import UnknownTarget
from avilla.core.ryanvk.collector.account import AccountCollector
from avilla.core.selector import Selector
from avilla.standard.core.privilege import MuteCapability, MuteInfo
from avilla.standard.core.profile import Nick, Summary

if TYPE_CHECKING:
    from avilla.red.account import RedAccount  # noqa
    from avilla.red.protocol import RedProtocol  # noqa


class RedMemberActionPerform((m := AccountCollector["RedProtocol", "RedAccount"]())._):
    m.namespace = "avilla.protocol/red::action"
    m.identify = "member"

    @m.pull("land.group.member", Summary)
    async def get_summary(self, target: Selector, route: ...) -> Summary:
        cache: Memcache = self.protocol.avilla.launch_manager.get_component(MemcacheService).cache
        if raw := await cache.get(
            f"red/account({self.account.route['account']}).group({target.pattern['group']}).member({target.pattern['member']})"
        ):
            return Summary(raw["nick"], "a member of this group")
        result = await self.account.websocket_client.call_http(
            "get", "api/group/getMemberList", {"group": target.pattern["group"]}
        )
        result = cast(list, result)
        for i in result:
            member_id = str(i["uin"])
            if member_id == target["member"]:
                return Summary(i["nick"], "name of member")
        raise UnknownTarget("Member not found")

    @m.pull("land.group.member", Nick)
    async def get_nick(self, target: Selector, route: ...) -> Nick:
        cache: Memcache = self.protocol.avilla.launch_manager.get_component(MemcacheService).cache
        if raw := await cache.get(
            f"red/account({self.account.route['account']}).group({target.pattern['group']}).member({target.pattern['member']})"
        ):
            return Nick(raw["nick"], raw["remark"] or raw["nick"], raw["cardName"])
        result = await self.account.websocket_client.call_http(
            "get", "api/group/getMemberList", {"group": target.pattern["group"]}
        )
        result = cast(list, result)
        for i in result:
            member_id = str(i["uin"])
            if member_id == target["member"]:
                return Nick(i["nick"], i["card"], i["cardName"])
        raise UnknownTarget("Member not found")

    @m.pull("land.group.member", MuteInfo)
    async def get_mute_info(self, target: Selector, route: ...) -> MuteInfo:
        cache: Memcache = self.protocol.avilla.launch_manager.get_component(MemcacheService).cache
        if raw := await cache.get(
            f"red/account({self.account.route['account']}).group({target.pattern['group']}).member({target.pattern['member']})"
        ):
            return MuteInfo(
                raw["shutUpTime"] > 0,
                timedelta(seconds=raw["shutUpTime"]),
            )
        result = await self.account.websocket_client.call_http(
            "get", "api/group/getMemberList", {"group": target.pattern["group"]}
        )
        result = cast(list, result)
        for i in result:
            member_id = str(i["uin"])
            if member_id == target["member"]:
                return MuteInfo(
                    i["shutUpTime"] > 0,
                    timedelta(seconds=i["shutUpTime"]),
                )
        raise UnknownTarget("Member not found")

    @m.entity(MuteCapability.mute, target="land.group.member")
    async def group_member_mute(self, target: Selector, duration: timedelta):
        time = max(60.0, min(duration.total_seconds(), 2592000))
        await self.account.websocket_client.call_http(
            "post",
            "api/group/muteMember",
            {"group": int(target["group"]), "memList": [{"uin": int(target["member"]), "timeStamp": time}]},
        )

    @m.entity(MuteCapability.unmute, target="land.group.member")
    async def group_member_unmute(self, target: Selector):
        await self.account.websocket_client.call_http(
            "post",
            "api/group/muteMember",
            {"group": int(target["group"]), "memList": [{"uin": int(target["member"]), "timeStamp": 0}]},
        )



from __future__ import annotations

from pathlib import Path

from avilla.core.context import Context
from avilla.core.resource import Resource
from avilla.core.selector import Selector


class RedResource(Resource[bytes]):
    id: str
    ctx: Context
    size: int
    name: str
    elem: str
    uuid: str

    def __init__(self, ctx: Context, selector: Selector, id: str, size: int, name: str, elem: str, uuid: str):
        super().__init__(selector)
        self.ctx = ctx
        self.id = id
        self.size = size
        self.name = name
        self.elem = elem
        self.uuid = uuid


class RedFileResource(RedResource):
    pass


class RedImageResource(RedResource):
    def __init__(
        self,
        ctx: Context,
        selector: Selector,
        id: str,
        size: int,
        name: str,
        elem: str,
        uuid: str,
        path: str | Path,
        width: int,
        height: int,
    ):
        super().__init__(ctx, selector, id, size, name, elem, uuid)
        self.path = Path(path)
        self.width = width
        self.height = height

    @property
    def url(self) -> str:
        if self.ctx.scene.last_key == "friend":
            return f"https://c2cpicdw.qpic.cn/offpic_new//{self.ctx.scene.last_value}-0-{self.id.upper()}/0"
        return f"https://gchat.qpic.cn/gchatpic_new/0/0-0-{self.id.upper()}/0"


class RedVoiceResource(RedResource):
    def __init__(
        self,
        ctx: Context,
        selector: Selector,
        id: str,
        size: int,
        name: str,
        elem: str,
        uuid: str,
        path: str | Path,
    ):
        super().__init__(ctx, selector, id, size, name, elem, uuid)
        self.path = Path(path)


class RedVideoResource(RedResource):
    def __init__(
        self,
        ctx: Context,
        selector: Selector,
        id: str,
        size: int,
        name: str,
        elem: str,
        uuid: str,
        path: str | Path,
    ):
        super().__init__(ctx, selector, id, size, name, elem, uuid)
        self.path = Path(path)



from __future__ import annotations

import random
from typing import TYPE_CHECKING

from graia.amnesia.builtins.memcache import Memcache, MemcacheService
from graia.amnesia.message import MessageChain
from loguru import logger

from avilla.core import Context
from avilla.core.ryanvk.collector.account import AccountCollector
from avilla.core.selector import Selector
from avilla.red.capability import RedCapability
from avilla.standard.core.message import (
    MessageReceived,
    MessageRevoke,
    MessageSend,
    MessageSent,
)
from avilla.standard.qq.elements import Forward, Node

if TYPE_CHECKING:
    from avilla.red.account import RedAccount  # noqa
    from avilla.red.protocol import RedProtocol  # noqa


class RedMessageActionPerform((m := AccountCollector["RedProtocol", "RedAccount"]())._):
    m.namespace = "avilla.protocol/red::action"
    m.identify = "message"

    async def handle_reply(self, target: Selector):
        cache: Memcache = self.protocol.avilla.launch_manager.get_component(MemcacheService).cache
        reply_msg = await cache.get(
            f"red/account({self.account.route['account']}).message({target.pattern['message']})"
        )
        if reply_msg:
            return {
                "elementType": 7,
                "replyElement": {
                    "replayMsgId": reply_msg["msgId"],
                    "replayMsgSeq": reply_msg["msgSeq"],
                    "senderUin": reply_msg["senderUin"],
                    "senderUinStr": str(reply_msg["senderUin"]),
                },
            }
        logger.warning(f"Unknown message {target.pattern['message']} for reply")
        return None

    @m.entity(MessageSend.send, target="land.group")
    async def send_group_msg(
        self,
        target: Selector,
        message: MessageChain,
        *,
        reply: Selector | None = None,
    ) -> Selector:
        if message.has(Forward):
            return await self.send_group_forward_msg(target, message.get_first(Forward))
        msg = await RedCapability(self.account.staff).serialize(message)
        if reply and (reply_msg := await self.handle_reply(reply)):
            msg.insert(0, reply_msg)
        resp = await self.account.websocket_client.call_http(
            "post",
            "api/message/send",
            {
                "peer": {
                    "chatType": 2,
                    "peerUin": target.pattern["group"],
                    "guildId": None,
                },
                "elements": msg,
            },
        )
        if "msgId" in resp:
            msg_id = resp["msgId"]
            event = await RedCapability(
                self.account.staff.ext({"connection": self.account.websocket_client})
            ).event_callback("message::recv", resp)
            if TYPE_CHECKING:
                assert isinstance(event, MessageReceived)
            event.context = self.account.get_context(target.member(self.account.route["account"]))
            event.message.scene = target
            event.message.sender = target.member(self.account.route["account"])
            self.protocol.post_event(MessageSent(event.context, event.message, self.account))
        else:
            msg_id = "unknown"
        return Selector().land(self.account.route["land"]).group(target.pattern["group"]).message(msg_id)

    @m.entity(MessageSend.send, target="land.friend")
    async def send_friend_msg(
        self,
        target: Selector,
        message: MessageChain,
        *,
        reply: Selector | None = None,
    ) -> Selector:
        msg = await RedCapability(self.account.staff).serialize(message)
        if reply and (reply_msg := await self.handle_reply(reply)):
            msg.insert(0, reply_msg)
        resp = await self.account.websocket_client.call_http(
            "post",
            "api/message/send",
            {
                "peer": {
                    "chatType": 1,
                    "peerUin": target.pattern["friend"],
                    "guildId": None,
                },
                "elements": msg,
            },
        )
        if "msgId" in resp:
            msg_id = resp["msgId"]
            event = await RedCapability(
                self.account.staff.ext({"connection": self.account.websocket_client})
            ).event_callback("message::recv", resp)
            if TYPE_CHECKING:
                assert isinstance(event, MessageReceived)
            event.context = Context(
                self.account,
                self.account.route,
                target,
                target,
                self.account.route,
            )
            event.message.scene = target
            event.message.sender = self.account.route
            self.protocol.post_event(MessageSent(event.context, event.message, self.account))
        else:
            msg_id = "unknown"
        return Selector().land(self.account.route["land"]).friend(target.pattern["friend"]).message(msg_id)

    @m.entity(MessageRevoke.revoke, target="land.group.message")
    async def revoke_group_msg(
        self,
        target: Selector,
    ) -> None:
        await self.account.websocket_client.call_http(
            "post",
            "api/message/recall",
            {
                "peer": {
                    "chatType": 2,
                    "peerUid": target.pattern["group"],
                    "guildId": None,
                },
                "msgId": [target.pattern["message"]],
            },
        )

    @m.entity(MessageRevoke.revoke, target="land.friend.message")
    async def revoke_friend_msg(
        self,
        target: Selector,
    ) -> None:
        await self.account.websocket_client.call_http(
            "post",
            "api/message/recall",
            {
                "peer": {
                    "chatType": 1,
                    "peerUid": target.pattern["friend"],
                    "guildId": None,
                },
                "msgId": [target.pattern["message"]],
            },
        )

    @m.entity(RedCapability.send_forward, target="land.group")
    async def send_group_forward_msg(self, target: Selector, forward: Forward) -> Selector:
        if all(node.mid for node in forward.nodes):
            await self.account.websocket_client.call_http(
                "post",
                "api/message/unsafeSendForward",
                {
                    "dstContact": {
                        "chatType": 2,
                        "peerUid": target.pattern["group"],
                        "guildId": None,
                    },
                    "srcContact": {
                        "chatType": 2,
                        "peerUid": target.pattern["group"],
                        "guildId": None,
                    },
                    "msgInfos": [
                        {"msgId": node.mid["message"], "senderShowName": node.name}  # type: ignore
                        for node in forward.nodes
                    ],
                },
            )
            return Selector().land(self.account.route["land"]).group(target.pattern["group"]).message("unknown")
        if all(node.content for node in forward.nodes):
            elems = []
            base_seq = random.randint(0, 65535)
            for node in forward.nodes:
                elems.append(await self.export_forward_node(base_seq, node, target))
                base_seq += 1
            await self.account.websocket_client.call_http(
                "post",
                "api/message/unsafeSendForward",
                {
                    "dstContact": {
                        "chatType": 2,
                        "peerUid": target.pattern["group"],
                        "guildId": None,
                    },
                    "srcContact": {
                        "chatType": 2,
                        "peerUid": target.pattern["group"],
                        "guildId": None,
                    },
                    "msgElements": elems,
                },
            )
            return Selector().land(self.account.route["land"]).group(target.pattern["group"]).message("unknown")
        raise ValueError("Forward message must have at least one node with content or mid")

    async def export_forward_node(self, seq: int, node: Node, target: Selector):
        cap = RedCapability(self.account.staff)
        elems = [await cap.forward_export(elem) for elem in node.content]  # type: ignore
        return {
            "head": {
                "field2": node.uid,
                "field8": {
                    "field1": target.pattern["group"],
                    "field4": node.name,
                },
            },
            "content": {
                "field1": 82,
                "field4": random.randint(0, 4294967295),
                "field5": seq,
                "field6": int(node.time.timestamp()),
                "field7": 1,
                "field8": 0,
                "field9": 0,
                "field15": {"field1": 0, "field2": 0},
            },
            "body": {"richText": {"elems": elems}},
        }



from __future__ import annotations

import random
from pathlib import Path
from typing import TYPE_CHECKING, Any

from avilla.core.builtins.resource_fetch import CoreResourceFetchPerform
from avilla.core.elements import Audio, Face, Notice, NoticeAll, Picture, Text
from avilla.core.resource import LocalFileResource, RawResource, UrlResource
from avilla.core.ryanvk.collector.account import AccountCollector
from avilla.red.capability import RedCapability
from avilla.standard.qq.elements import MarketFace

if TYPE_CHECKING:
    from avilla.red.account import RedAccount  # noqa
    from avilla.red.protocol import RedProtocol  # noqa


class RedMessageSerializePerform((m := AccountCollector["RedProtocol", "RedAccount"]())._):
    m.namespace = "avilla.protocol/red::message"
    m.identify = "serialize"

    # LINK: https://github.com/microsoft/pyright/issues/5409

    @m.entity(RedCapability.serialize_element, element=Text)
    async def text(self, element: Text) -> dict:
        return {"elementType": 1, "textElement": {"content": element.text}}

    @m.entity(RedCapability.serialize_element, element=Face)
    async def face(self, element: Face) -> dict:
        return {"elementType": 6, "faceElement": {"faceIndex": element.id}}

    @m.entity(RedCapability.serialize_element, element=Notice)
    async def notice(self, element: Notice) -> dict:
        return {
            "elementType": 1,
            "textElement": {
                "atType": 2,
                "atNtUin": element.target.last_value,
                "content": f"@{element.display or element.target.last_value}",
            },
        }

    @m.entity(RedCapability.serialize_element, element=NoticeAll)
    async def notice_all(self, element: NoticeAll) -> dict:
        return {"elementType": 1, "textElement": {"atType": 1}}

    @m.entity(RedCapability.serialize_element, element=Picture)
    async def picture(self, element: Picture) -> dict:
        if isinstance(element.resource, RawResource):
            data = await CoreResourceFetchPerform(self.account.staff).fetch_raw(element.resource)
        elif isinstance(element.resource, UrlResource):
            data = await CoreResourceFetchPerform(self.account.staff).fetch_url(element.resource)
        elif isinstance(element.resource, LocalFileResource):
            data = await CoreResourceFetchPerform(self.account.staff).fetch_localfile(element.resource)
        else:
            data = await self.account.staff.fetch_resource(element.resource)
        resp = await self.account.websocket_client.call_http(
            "multipart",
            "api/upload",
            {
                "file": {
                    "value": data,
                    "content_type": None,
                    "filename": "file_image",
                }
            },
        )
        file = Path(resp["ntFilePath"])
        return {
            "elementType": 2,
            "picElement": {
                "original": True,
                "md5HexStr": resp["md5"],
                "picWidth": resp["imageInfo"]["width"],
                "picHeight": resp["imageInfo"]["height"],
                "fileSize": resp["fileSize"],
                "fileName": file.name,
                "sourcePath": resp["ntFilePath"],
            },
        }

    @m.entity(RedCapability.serialize_element, element=MarketFace)
    async def market_face(self, element: MarketFace) -> dict:
        emoji_id, key, emoji_package_id = element.id.split("/")
        return {
            "elementType": 11,
            "marketFaceElement": {
                "emojiId": int(emoji_id),
                "key": key,
                "emojiPackageId": int(emoji_package_id),
            },
        }

    @m.entity(RedCapability.serialize_element, element=Audio)
    async def audio(self, element: Audio) -> dict:
        if isinstance(element.resource, RawResource):
            data = await CoreResourceFetchPerform(self.account.staff).fetch_raw(element.resource)
        elif isinstance(element.resource, UrlResource):
            data = await CoreResourceFetchPerform(self.account.staff).fetch_url(element.resource)
        elif isinstance(element.resource, LocalFileResource):
            data = await CoreResourceFetchPerform(self.account.staff).fetch_localfile(element.resource)
        else:
            data = await self.account.staff.fetch_resource(element.resource)
        resp = await self.account.websocket_client.call_http(
            "multipart",
            "api/upload",
            {
                "file": {
                    "value": data,
                    "content_type": None,
                    "filename": "file_audio",
                }
            },
        )
        file = Path(resp["ntFilePath"])
        return {
            "elementType": 4,
            "pttElement": {
                "canConvert2Text": True,
                "md5HexStr": resp["md5"],
                "fileSize": resp["fileSize"],
                "fileName": file.name,
                "filePath": resp["ntFilePath"],
                "duration": max(1, element.duration),
                "formatType": 1,
                "voiceType": 1,
                "voiceChangeType": 0,
                "playState": 1,
                "waveAmplitudes": [99 for _ in range(17)],
            },
        }

    @m.entity(RedCapability.forward_export, element=Text)
    async def forward_text(self, element: Text) -> dict:
        return {"text": {"str": element.text}}

    @m.entity(RedCapability.forward_export, element=Notice)
    async def forward_notice(self, element: Notice) -> dict:
        return {"text": {"str": f"@{element.display or element.target.last_value}"}}

    @m.entity(RedCapability.forward_export, element=NoticeAll)
    async def forward_notice_all(self, element: NoticeAll) -> dict:
        return {"text": {"str": "@全体成员"}}

    @m.entity(RedCapability.forward_export, element=Picture)
    async def forward_picture(self, element: Picture) -> dict:
        if isinstance(element.resource, RawResource):
            data = await CoreResourceFetchPerform(self.account.staff).fetch_raw(element.resource)
        elif isinstance(element.resource, UrlResource):
            data = await CoreResourceFetchPerform(self.account.staff).fetch_url(element.resource)
        elif isinstance(element.resource, LocalFileResource):
            data = await CoreResourceFetchPerform(self.account.staff).fetch_localfile(element.resource)
        else:
            data = await self.account.staff.fetch_resource(element.resource)
        resp = await self.account.websocket_client.call_http(
            "multipart",
            "api/upload",
            {
                "file": {
                    "value": data,
                    "content_type": None,
                    "filename": "file_image",
                }
            },
        )
        md5 = resp["md5"]
        file = Path(resp["ntFilePath"])
        pid = f"{{{md5[:8].upper()}-{md5[8:12].upper()}-{md5[12:16].upper()}-{md5[16:20].upper()}-{md5[20:].upper()}}}{file.suffix}"  # noqa: E501
        return {
            "customFace": {
                "filePath": pid,
                "fileId": random.randint(0, 65535),
                "serverIp": -1740138629,
                "serverPort": 80,
                "fileType": 1001,
                "useful": 1,
                "md5": [int(md5[i : i + 2], 16) for i in range(0, 32, 2)],
                "imageType": 1001,
                "width": resp["imageInfo"]["width"],
                "height": resp["imageInfo"]["height"],
                "size": resp["fileSize"],
                "origin": 0,
                "thumbWidth": 0,
                "thumbHeight": 0,
                "pbReserve": [2, 0],
            }
        }

    @m.entity(RedCapability.forward_export, element=Any)
    async def forward_any(self, element: Any) -> dict:
        return {"text": {"str": str(element)}}



from __future__ import annotations

from loguru import logger
from selectolax.parser import HTMLParser

from avilla.core.context import Context
from avilla.core.event import MemberCreated
from avilla.core.selector import Selector
from avilla.red.capability import RedCapability
from avilla.red.collector.connection import ConnectionCollector

# async function adaptGuildMemberAddedMessage(
#   session: Session,
#   data: Message,
# ): Promise<Session | undefined> {
#   const session2 = await adaptChatMessage(session, data)
#   session2.type = 'guild-member-added'
#
#   session2.operatorId =
#     data.elements[0]!.grayTipElement!.groupElement!.adminUin!
#   session2.author = {
#     userId: data.elements[0]!.grayTipElement!.groupElement!.memberUin!,
#     username: data.elements[0]!.grayTipElement!.groupElement!.memberNick!,
#     nickname: data.elements[0]!.grayTipElement!.groupElement!.memberNick!,
#   }
#   session2.userId = session2.author.userId
#
#   return session2
# }


class RedEventRelationshipPerform((m := ConnectionCollector())._):
    m.namespace = "avilla.protocol/red::event"
    m.identify = "relationship"

    @m.entity(RedCapability.event_callback, event_type="group::member::add")
    async def member_join(self, event_type: ..., raw_event: dict):
        account = self.connection.account
        if account is None:
            logger.warning(f"Unknown account received message {raw_event}")
            return
        group = Selector().land(account.route["land"]).group(str(raw_event.get("peerUin", raw_event.get("peerUid"))))
        group_data = raw_event["elements"][0]["grayTipElement"]["groupElement"]
        member = group.member(str(group_data["memberUin"]))
        operator = group.member(str(group_data["adminUin"]))
        context = Context(account, member, group, group, group.member(account.route["account"]), mediums=[operator])
        return MemberCreated(context)

    @m.entity(RedCapability.event_callback, event_type="group::member::legacy::add::invited")
    async def member_join_invited(self, event_type: ..., raw_event: dict):
        account = self.connection.account
        if account is None:
            logger.warning(f"Unknown account received message {raw_event}")
            return
        group = Selector().land(account.route["land"]).group(str(raw_event.get("peerUin", raw_event.get("peerUid"))))
        group_data = raw_event["elements"][0]["grayTipElement"]["xmlElement"]
        root = HTMLParser(group_data["content"])
        operator = group.member(root.tags("qq")[0].attributes["jp"])  # type: ignore
        member = group.member(root.tags("qq")[1].attributes["jp"])  # type: ignore
        context = Context(account, member, group, group, group.member(account.route["account"]), mediums=[operator])
        return MemberCreated(context)



from __future__ import annotations

from datetime import timedelta
from typing import TYPE_CHECKING, Callable, cast

from graia.amnesia.builtins.memcache import Memcache, MemcacheService

from avilla.core.builtins.capability import CoreCapability
from avilla.core.ryanvk.collector.account import AccountCollector
from avilla.core.selector import Selector

if TYPE_CHECKING:
    from ..account import RedAccount  # noqa
    from ..protocol import RedProtocol  # noqa


class RedQueryPerform((m := AccountCollector["RedProtocol", "RedAccount"]())._):
    m.namespace = "avilla.protocol/red::query"

    @m.entity(CoreCapability.query, target="land.group")
    async def query_group(self, predicate: Callable[[str, str], bool] | str, previous: None):
        cache: Memcache = self.protocol.avilla.launch_manager.get_component(MemcacheService).cache
        result = await self.account.websocket_client.call_http("get", "api/bot/groups", {})
        result = cast(list, result)
        for i in result:
            group_id = str(i["groupCode"])
            await cache.set(f"red/account({self.account.route['account']}).group({group_id})", i, timedelta(minutes=5))
            if callable(predicate) and predicate("group", group_id) or group_id == predicate:
                yield Selector().land(self.account.route["land"]).group(group_id)

    @m.entity(CoreCapability.query, target="land.friend")
    async def query_friend(self, predicate: Callable[[str, str], bool] | str, previous: None):
        cache: Memcache = self.protocol.avilla.launch_manager.get_component(MemcacheService).cache
        result = await self.account.websocket_client.call_http("get", "api/bot/friends", {})
        result = cast(list, result)
        for i in result:
            friend_id = str(i["uin"])
            await cache.set(
                f"red/account({self.account.route['account']}).friend({friend_id})", i, timedelta(minutes=5)
            )
            if callable(predicate) and predicate("friend", friend_id) or friend_id == predicate:
                yield Selector().land(self.account.route["land"]).friend(friend_id)

    @m.entity(CoreCapability.query, target="member", previous="land.group")  # type: ignore
    async def query_group_members(self, predicate: Callable[[str, str], bool] | str, previous: Selector):
        cache: Memcache = self.protocol.avilla.launch_manager.get_component(MemcacheService).cache
        result = await self.account.websocket_client.call_http(
            "post", "api/group/getMemberList", {"group": int(previous["group"])}
        )
        result = cast(list, result)
        for i in result:
            member_id = str(i["uin"])
            await cache.set(
                f"red/account({self.account.route['account']}).group({previous['group']}).member({member_id})",
                i,
                timedelta(minutes=5),
            )
            if callable(predicate) and predicate("member", member_id) or member_id == predicate:
                yield previous.member(member_id)



from __future__ import annotations

from dataclasses import InitVar, dataclass, field

from yarl import URL

from avilla.core.application import Avilla
from avilla.core.protocol import BaseProtocol, ProtocolConfig
from graia.ryanvk import merge, ref

from .net.ws_client import RedWsClientNetworking
from .service import RedService


@dataclass
class RedConfig(ProtocolConfig):
    access_token: str
    host: str = field(default="localhost")
    port: int = field(default=16530)
    _http_host: InitVar[str | None] = None
    endpoint: URL = field(init=False)
    http_endpoint: URL = field(init=False)

    def __post_init__(self, _http_host: str | None):
        if not self.access_token:
            raise ValueError("access_token must be set")
        self.endpoint = URL.build(scheme="ws", host=self.host, port=self.port)
        self.http_endpoint = URL.build(scheme="http", host=_http_host or self.host, port=self.port)


def _import_performs():  # noqa: F401
    # isort: off

    # :: Message
    import avilla.red.perform.message.deserialize  # noqa: F401
    import avilla.red.perform.message.serialize  # noqa: F401

    ## :: Action
    import avilla.red.perform.action.friend  # noqa: F401
    import avilla.red.perform.action.group  # noqa: F401
    import avilla.red.perform.action.member  # noqa: F401
    import avilla.red.perform.action.message  # noqa: F401

    ## :: Context
    import avilla.red.perform.context  # noqa: F401

    ## :: Event
    import avilla.red.perform.event.group  # noqa: F401
    import avilla.red.perform.event.member  # noqa: F401
    import avilla.red.perform.event.message  # noqa: F401
    import avilla.red.perform.event.lifespan  # noqa: F401
    import avilla.red.perform.event.relationship  # noqa: F401

    ## :: Query
    import avilla.red.perform.query  # noqa: F401

    ## :: Resource Fetch
    import avilla.red.perform.resource_fetch  # noqa: F401


class RedProtocol(BaseProtocol):
    service: RedService

    _import_performs()
    artifacts = {
        **merge(
            ref("avilla.protocol/red::context"),
            ref("avilla.protocol/red::query"),
            ref("avilla.protocol/red::resource_fetch"),
            ref("avilla.protocol/red::action", "message"),
            ref("avilla.protocol/red::action", "friend"),
            ref("avilla.protocol/red::action", "group"),
            ref("avilla.protocol/red::action", "member"),
            ref("avilla.protocol/red::message", "deserialize"),
            ref("avilla.protocol/red::message", "serialize"),
            ref("avilla.protocol/red::event", "message"),
            ref("avilla.protocol/red::event", "lifespan"),
            ref("avilla.protocol/red::event", "relationship"),
            ref("avilla.protocol/red::event", "group"),
            ref("avilla.protocol/red::event", "member"),
        ),
    }

    def __init__(self):
        self.service = RedService(self)

    def ensure(self, avilla: Avilla):
        self.avilla = avilla

        avilla.launch_manager.add_component(self.service)

    def configure(self, config: RedConfig):
        self.service.connections.append(RedWsClientNetworking(self, config))
        return self



from __future__ import annotations

from typing import TYPE_CHECKING

from graia.amnesia.message import MessageChain
from graia.amnesia.message.element import Unknown
from selectolax.parser import HTMLParser

from avilla.core.elements import (
    Audio,
    Face,
    File,
    Notice,
    NoticeAll,
    Picture,
    Text,
    Video,
)
from avilla.core.ryanvk.collector.application import ApplicationCollector
from avilla.core.selector import Selector
from avilla.red.capability import RedCapability
from avilla.red.resource import (
    RedFileResource,
    RedImageResource,
    RedVideoResource,
    RedVoiceResource,
)
from avilla.standard.qq.elements import (
    App,
    DisplayStrategy,
    Forward,
    MarketFace,
    Node,
    Poke,
    PokeKind,
)
from graia.ryanvk import OptionalAccess

if TYPE_CHECKING:
    from avilla.core.context import Context
    from avilla.red.account import RedAccount


class RedMessageDeserializePerform((m := ApplicationCollector())._):
    m.namespace = "avilla.protocol/red::message"
    m.identify = "deserialize"

    # LINK: https://github.com/microsoft/pyright/issues/5409
    context: OptionalAccess[Context] = OptionalAccess()
    account: OptionalAccess[RedAccount] = OptionalAccess()

    @m.entity(RedCapability.deserialize_element, element="text")
    async def text(self, element: dict) -> Text | Notice | NoticeAll:
        if not element["atType"]:
            return Text(element["content"])
        if element["atType"] == 1:
            return NoticeAll()
        if self.context:
            return Notice(
                self.context.scene.member(element.get("atNtUin", "atNtUid")),
                element["content"][1:],
            )
        return Notice(
            Selector().land("qq").member(element.get("atNtUin", "atNtUid")),
            element["content"][1:],
        )

    @m.entity(RedCapability.deserialize_element, element="face")
    async def face(self, element: dict) -> Face | Poke:
        if element["faceType"] == 5:
            return Poke(PokeKind.ChuoYiChuo)
        return Face(element["faceIndex"], element["faceText"])

    @m.entity(RedCapability.deserialize_element, element="pic")
    async def pic(self, element: dict) -> Picture:
        resource = RedImageResource(
            self.context,  # type: ignore
            Selector().land("qq").picture(md5 := element["md5HexStr"]),
            md5,
            element["fileSize"],
            element["fileName"],
            element["elementId"],
            element["fileUuid"],
            element["sourcePath"],
            element["picWidth"],
            element["picHeight"],
        )
        return Picture(resource)

    @m.entity(RedCapability.deserialize_element, element="marketFace")
    async def market_face(self, element: dict) -> MarketFace:
        return MarketFace(
            f"{element['emojiId']}/{element['key']}/{element['emojiPackageId']}",
        )

    @m.entity(RedCapability.deserialize_element, element="ark")
    async def ark(self, element: dict) -> App:
        return App(element["bytesData"])

    @m.entity(RedCapability.deserialize_element, element="file")
    async def file(self, element: dict) -> File:
        return File(
            RedFileResource(
                self.context,  # type: ignore
                Selector().land("qq").file(element["fileMd5"]),
                element["fileMd5"],
                element["fileSize"],
                element["fileName"],
                element["elementId"],
                element["fileUuid"],
            )
        )

    @m.entity(RedCapability.deserialize_element, element="ptt")
    async def ptt(self, element: dict) -> Audio:
        return Audio(
            RedVoiceResource(
                self.context,  # type: ignore
                Selector().land("qq").voice(element["md5HexStr"]),
                element["md5HexStr"],
                element.get("fileSize", 0),
                element["fileName"],
                element["elementId"],
                element["fileUuid"],
                element["filePath"],
            ),
            int(element["duration"]),
        )

    @m.entity(RedCapability.deserialize_element, element="grayTip")
    async def gray_tip(self, element: dict) -> Unknown:
        return Unknown("grayTip", element)

    @m.entity(RedCapability.deserialize_element, element="multiForwardMsg")
    async def forward(self, element: dict) -> Forward:
        root = HTMLParser(element["xmlContent"])
        title = root.css_first("source").attributes["name"]
        summary = root.css_first("summary").text()
        brief = root.css_first("msg").attributes["brief"]
        preview = [node.text() for node in root.tags("title")[1:]]
        return Forward(
            element["resId"],
            nodes=[
                Node(name=(part := content.split(":", 1))[0], content=MessageChain([Text(part[1].lstrip())]))
                for content in preview
            ],
            strategy=DisplayStrategy(title, brief, preview=preview, summary=summary),
        )

    @m.entity(RedCapability.deserialize_element, element="video")
    async def video(self, element: dict) -> Video:
        return Video(
            RedVideoResource(
                self.context,  # type: ignore
                Selector().land("qq").video(element["videoMd5"]),
                element["videoMd5"],
                element["fileSize"],
                element["fileName"],
                element["elementId"],
                element["fileUuid"],
                element["filePath"],
            )
        )



from __future__ import annotations

from typing import Any

from graia.amnesia.message import Element, MessageChain

from avilla.core.event import AvillaEvent
from avilla.core.ryanvk.collector.application import ApplicationCollector
from avilla.core.ryanvk.overload.target import TargetOverload
from avilla.core.selector import Selector
from avilla.standard.core.application.event import AvillaLifecycleEvent
from avilla.standard.qq.elements import Forward
from graia.ryanvk import Fn, PredicateOverload, SimpleOverload, TypeOverload


class RedCapability((m := ApplicationCollector())._):
    @Fn.complex({SimpleOverload(): ["event_type"]})
    async def event_callback(self, event_type: str, raw_event: dict) -> AvillaEvent | AvillaLifecycleEvent | None:
        ...

    @Fn.complex({PredicateOverload(lambda _, raw: raw["type"]): ["element"]})
    async def deserialize_element(self, element: dict) -> Element:  # type: ignore
        ...

    @Fn.complex({TypeOverload(): ["element"]})
    async def serialize_element(self, element: Any) -> dict:  # type: ignore
        ...

    @Fn.complex({TypeOverload(): ["element"]})
    async def forward_export(self, element: Any) -> dict:  # type: ignore
        ...

    @Fn.complex({TargetOverload(): ["target"]})
    async def send_forward(self, target: Selector, forward: Forward) -> Selector:
        ...

    async def deserialize(self, elements: list[dict]):
        _elements = []

        for raw_element in elements:
            _elements.append(await self.deserialize_element(raw_element))

        return MessageChain(_elements)

    async def serialize(self, message: MessageChain):
        chain = []

        for element in message:
            chain.append(await self.serialize_element(element))

        return chain

    async def handle_event(self, event_type: str, payload: dict):
        maybe_event = await self.event_callback(event_type, payload)

        if maybe_event is not None:
            self.avilla.event_record(maybe_event)
            self.avilla.broadcast.postEvent(maybe_event)



from __future__ import annotations

import asyncio
from contextlib import suppress
from typing import TYPE_CHECKING, AsyncIterator, Literal, overload

from loguru import logger
from typing_extensions import Self

from avilla.core.ryanvk.staff import Staff
from avilla.red.account import RedAccount
from avilla.red.capability import RedCapability
from avilla.red.utils import MsgType, get_msg_types

if TYPE_CHECKING:
    from avilla.red.protocol import RedProtocol


class RedNetworking:
    protocol: RedProtocol
    account: RedAccount | None
    close_signal: asyncio.Event

    def __init__(self, protocol: RedProtocol):
        super().__init__()
        self.protocol = protocol
        self.account = None
        self.close_signal = asyncio.Event()

    def get_staff_components(self):
        return {"connection": self, "protocol": self.protocol, "avilla": self.protocol.avilla}

    def get_staff_artifacts(self):
        return [self.protocol.artifacts, self.protocol.avilla.global_artifacts]

    @property
    def staff(self):
        return Staff(self.get_staff_artifacts(), self.get_staff_components())

    def message_receive(self) -> AsyncIterator[tuple[Self, dict]]:
        ...

    @property
    def alive(self) -> bool:
        ...

    async def wait_for_available(self):
        ...

    async def send(self, payload: dict) -> None:
        ...

    async def message_handle(self):
        async for connection, data in self.message_receive():
            event_type = data["type"]
            if not data["payload"]:
                logger.warning(f"received empty event {event_type}")
                continue

            async def event_parse_task(t: str, payload: dict):
                with suppress(NotImplementedError):
                    await RedCapability(connection.staff).handle_event(t, payload)
                    return
                logger.warning(f"received unsupported event {t}: {payload}")

            def handle_message(message: dict):
                types = get_msg_types(message)
                if types.msg == MsgType.system and types.send == "system":
                    if (
                        message["subMsgType"] == 8
                        and message["elements"][0]["elementType"] == 8
                        and message["elements"][0]["grayTipElement"]["subElementType"] == 4
                        and message["elements"][0]["grayTipElement"]["groupElement"]["type"] == 1
                    ):
                        asyncio.create_task(event_parse_task("group::member::add", message))
                    elif (
                        message["subMsgType"] == 8
                        and message["elements"][0]["elementType"] == 8
                        and message["elements"][0]["grayTipElement"]["subElementType"] == 4
                        and message["elements"][0]["grayTipElement"]["groupElement"]["type"] == 8
                    ):
                        asyncio.create_task(event_parse_task("group::member::mute", message))
                    elif (
                        message["subMsgType"] == 8
                        and message["elements"][0]["elementType"] == 8
                        and message["elements"][0]["grayTipElement"]["subElementType"] == 4
                        and message["elements"][0]["grayTipElement"]["groupElement"]["type"] == 5
                    ):
                        asyncio.create_task(event_parse_task("group::name_update", message))
                    elif (
                        message["subMsgType"] == 12
                        and message["elements"][0]["elementType"] == 8
                        and message["elements"][0]["grayTipElement"]["subElementType"] == 12
                        and message["elements"][0]["grayTipElement"]["xmlElement"]["busiType"] == "1"
                        and message["elements"][0]["grayTipElement"]["xmlElement"]["busiId"] == "10145"
                    ):
                        asyncio.create_task(event_parse_task("group::member::legacy::add::invited", message))
                    else:
                        logger.warning(f"received unsupported event: {message}")
                        return
                else:
                    asyncio.create_task(event_parse_task("message::recv", message))

            if event_type == "message::recv":
                for msg in data["payload"]:
                    handle_message(msg)
            else:
                asyncio.create_task(event_parse_task(event_type, data["payload"]))

    async def connection_closed(self):
        self.close_signal.set()

    async def call(self, action: str, params: dict | None = None) -> None:
        if not self.alive:
            raise RuntimeError("connection is not established")

        await self.wait_for_available()
        await self.send({"type": action, "payload": params or {}})
        return

    @overload
    async def call_http(
        self, method: Literal["get", "post", "multipart"], action: str, params: dict | None = None
    ) -> dict:
        ...

    @overload
    async def call_http(
        self,
        method: Literal["get", "post", "multipart"],
        action: str,
        params: dict | None = None,
        raw: Literal[True] = True,
    ) -> bytes:
        ...

    async def call_http(
        self, method: Literal["get", "post", "multipart"], action: str, params: dict | None = None, raw: bool = False
    ) -> dict | bytes:
        ...



from __future__ import annotations

from datetime import datetime, timedelta

from graia.amnesia.builtins.memcache import Memcache, MemcacheService
from loguru import logger

from avilla.core.context import Context
from avilla.core.message import Message
from avilla.core.selector import Selector
from avilla.red.capability import RedCapability
from avilla.red.collector.connection import ConnectionCollector
from avilla.red.utils import pre_deserialize
from avilla.standard.core.message import MessageReceived, MessageSent


class RedEventMessagePerform((m := ConnectionCollector())._):
    m.namespace = "avilla.protocol/red::event"
    m.identify = "message"

    @m.entity(RedCapability.event_callback, event_type="message::recv")
    async def message(self, event_type: ..., raw_event: dict):
        account = self.connection.account
        if account is None:
            logger.warning(f"Unknown account received message {raw_event}")
            return
        cache: Memcache = self.protocol.avilla.launch_manager.get_component(MemcacheService).cache
        reply = None
        if raw_event["chatType"] == 2:
            group = (
                Selector().land(account.route["land"]).group(str(raw_event.get("peerUin", raw_event.get("peerUid"))))
            )
            member = group.member(str(raw_event.get("senderUin", raw_event.get("senderUid"))))
            context = Context(
                account,
                member,
                group,
                group,
                group.member(account.route["account"]),
            )
            elements = pre_deserialize(raw_event["elements"])
            if elements[0]["type"] == "reply":
                reply = group.message(f"{elements[0]['sourceMsgIdInRecords']}")
                elements = elements[1:]
            message = await RedCapability(account.staff.ext({"context": context})).deserialize(elements)
            msg = Message(
                id=f'{raw_event["msgId"]}',
                scene=group,
                sender=member,
                content=message,
                time=datetime.fromtimestamp(int(raw_event["msgTime"])),
                reply=reply,
            )
        else:
            friend = (
                Selector().land(account.route["land"]).friend(f"{raw_event.get('peerUin', raw_event.get('senderUin'))}")
            )
            context = Context(
                account,
                friend,
                account.route,
                friend,
                account.route,
            )
            elements = pre_deserialize(raw_event["elements"])
            if elements[0]["type"] == "reply":
                reply = friend.message(f"{elements[0]['sourceMsgIdInRecords']}")
                elements = elements[1:]
            message = await RedCapability(account.staff.ext({"context": context})).deserialize(elements)
            msg = Message(
                id=f'{raw_event["msgId"]}',
                scene=friend,
                sender=friend,
                content=message,
                time=datetime.fromtimestamp(int(raw_event["msgTime"])),
                reply=reply,
            )
        await cache.set(f"red/account({account.route['account']}).message({msg.id})", raw_event, timedelta(minutes=5))
        context._collect_metadatas(msg.to_selector(), msg)
        return (
            MessageSent(context, msg, account)
            if msg.sender.last_value == account.route["account"]
            else MessageReceived(context, msg)
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Set

from launart import Launart, Service, any_completed

from .net.ws_client import RedWsClientNetworking

if TYPE_CHECKING:
    from .protocol import RedProtocol


class RedService(Service):
    id = "red.service"

    protocol: RedProtocol
    connections: list[RedWsClientNetworking]

    def __init__(self, protocol: RedProtocol):
        self.protocol = protocol
        self.connections = []
        super().__init__()

    async def launch(self, manager: Launart):
        async with self.stage("preparing"):
            for i in self.connections:
                manager.add_component(i)

        async with self.stage("blocking"):
            await any_completed(
                manager.status.wait_for_sigexit(), *(i.status.wait_for("blocking-completed") for i in self.connections)
            )

        async with self.stage("cleanup"):
            ...

    @property
    def stages(self):
        return {"preparing", "blocking", "cleanup"}

    @property
    def required(self) -> Set[str]:
        return set()



from __future__ import annotations

from typing import TYPE_CHECKING

from avilla.core.builtins.capability import CoreCapability
from avilla.core.context import Context
from avilla.core.ryanvk.collector.account import AccountCollector
from avilla.core.selector import Selector

if TYPE_CHECKING:
    from avilla.red.account import RedAccount  # noqa
    from avilla.red.protocol import RedProtocol  # noqa


class RedContextPerform((m := AccountCollector["RedProtocol", "RedAccount"]())._):
    m.namespace = "avilla.protocol/red::context"

    @m.entity(CoreCapability.get_context, target="land.group")
    def get_context_from_group(self, target: Selector, *, via: Selector | None = None):
        return Context(
            self.account,
            target,
            target,
            target,
            target.member(self.account.route["account"]),
        )

    @m.entity(CoreCapability.get_context, target="land.friend")
    def get_context_from_friend(self, target: Selector, *, via: Selector | None = None):
        if via:
            return Context(
                self.account,
                via,
                target,
                target,
                self.account.route,
            )
        return Context(self.account, target, self.account.route, target, self.account.route)

    @m.entity(CoreCapability.get_context, target="land.group.member")
    def get_context_from_member(self, target: Selector, *, via: Selector | None = None):
        return Context(
            self.account,
            target,
            target.into("::group"),
            target.into("::group"),
            target.into(f"~.member({self.account.route['account']})"),
        )

    @m.entity(CoreCapability.channel, target="land.group")
    @m.entity(CoreCapability.channel, target="land.group.member")
    def channel_from_group(self, target: Selector):
        return target["group"]

    @m.entity(CoreCapability.guild, target="land.group")
    @m.entity(CoreCapability.guild, target="land.group.member")
    def guild_from_group(self, target: Selector):
        return target["group"]

    @m.entity(CoreCapability.user, target="land.group.member")
    def user_from_member(self, target: Selector):
        return target["member"]

    @m.entity(CoreCapability.user, target="land.friend")
    @m.entity(CoreCapability.channel, target="land.friend")
    def user_from_friend(self, target: Selector):
        return target["friend"]



from dataclasses import dataclass
from enum import IntEnum
from typing import Literal, TypedDict


class MsgType(IntEnum):
    normal = 2
    may_file = 3
    system = 5
    voice = 6
    video = 7
    value8 = 8
    reply = 9
    wallet = 10
    ark = 11
    may_market = 17


class SubMsgTypes(TypedDict):
    text: bool
    image: bool
    face: bool
    link: bool
    forward: bool
    reply: bool
    market_face: bool
    file: bool


@dataclass
class MsgTypes:
    chat: Literal["friend", "group"]
    msg: MsgType
    sub: SubMsgTypes
    send: Literal["system", "normal"]

    @property
    def group(self):
        return self.chat == "group"


def pre_deserialize(elements: list[dict]):
    res = []
    for elem in elements:
        slot = {}
        data = {k: v for k, v in elem.items() if v is not None}
        for k, v in data.items():
            if k.endswith("Element"):
                slot["type"] = k[:-7]
                slot.update(v)
            elif k != "elementType":
                slot[k] = v
        res.append(slot)
    return res


def get_msg_types(raw_event: dict) -> MsgTypes:
    return MsgTypes(
        **{
            "chat": "friend" if raw_event["chatType"] == 1 else "group",
            "msg": MsgType(raw_event["msgType"]),
            "sub": {
                "text": bool(raw_event["subMsgType"] & (1 << 0)),
                "image": bool(raw_event["subMsgType"] & (1 << 1)),
                "face": bool(raw_event["subMsgType"] & (1 << 4)),
                "link": bool(raw_event["subMsgType"] & (1 << 7)),
                "forward": bool(raw_event["subMsgType"] & (1 << 3)),
                "reply": (raw_event["msgType"] == 9 and bool(raw_event["subMsgType"] & (1 << 5))),
                "market_face": (raw_event["msgType"] == 17 and bool(raw_event["subMsgType"] & (1 << 3))),
                "file": (raw_event["msgType"] == 3 and bool(raw_event["subMsgType"] & (1 << 9))),
            },
            "send": "system" if raw_event["sendType"] == 3 else "normal",
        }
    )


"""\
1 ===> text, at, ...
    content
    atType {0: not At, 1: everyone, 2: someone (atNtUid, atNtUin)}
2 ===> image(pic)
    fileName
    md5HexStr
    sourcePath
    fileUuid
    fileSubId
    picSubType {0: normal image (summary = ''), 1: emoji (summary = '[动画表情]')}
3 ===> file
    fileMd5
    fileName
    fileSize
    fileUuid
4 ===> voice(ptt)
    fileName
    filePath
    md5HexStr
    voiceChangeType {0: normal, 1: magic}
    canConvert2Text
        text
    waveAmplitudes
    fileUuid
5 ===> video
    filePath
    fileName
    videoMd5
    thumbMd5
    fileTime
    thumbSize
    fileFormat
    fileSize
    thumbWidth
    thumbHeight
    busiType
    subBusiType
    thumbPath
    transferStatus
    progress
    invalidState
    fileUuid
    fileSubId
    fileBizId
6 ===> face, poke
    faceIndex
    faceText {None: normal, '/xxx': sticker, '': poke}
    faceType {1: normal, 2: normal-extended, 3: sticker, 5: poke}
    packId {None: other, '1': sticker}
    stickerId {None: other, 'xxx': sticker}
    sourceType {None: other, 1: sticker}
    stickerType {None: other, 1: sticker}
    randomType {None: other, 1: sticker}
    pokeType {None: other, xxx: poke}
    spokeSummary {None: other, '': poke}
    doubleHit {None: other, xxx: poke}
    vaspokeId {None: other, xxx: poke}
    vaspokeName {None: other, 'xxx': poke}
7 ===> reply
    replayMsgId
    replayMsgSeq
    replyMsgTime
    sourceMsgIdInRecords
    sourceMsgTextElems
    senderUid
    senderUidStr
    senderUin
8 ==> grayTip
10 ===> app(ark) # 小程序，公告什么的
    bytesData (application/json)
11 ===> marketFace
    itemType
    faceInfo
    emojiPackageId
    subType
    faceName
    emojiId
    key
    staticFacePath
    dynamicFacePath
16 ===> multiForwardMsg
    xmlContent
    resId
    fileName
"""



from __future__ import annotations

from contextlib import suppress
from typing import TYPE_CHECKING

from aiohttp import ClientSession

from avilla.core.builtins.capability import CoreCapability
from avilla.core.ryanvk.collector.protocol import ProtocolCollector
from avilla.red.resource import (
    RedFileResource,
    RedImageResource,
    RedResource,
    RedVideoResource,
    RedVoiceResource,
)

if TYPE_CHECKING:
    from ..account import RedAccount  # noqa
    from ..protocol import RedProtocol  # noqa


class RedResourceFetchPerform((m := ProtocolCollector["RedProtocol"]())._):
    m.namespace = "avilla.protocol/red::resource_fetch"

    @m.entity(CoreCapability.fetch, resource=RedResource)
    @m.entity(CoreCapability.fetch, resource=RedFileResource)
    @m.entity(CoreCapability.fetch, resource=RedImageResource)
    @m.entity(CoreCapability.fetch, resource=RedVoiceResource)
    @m.entity(CoreCapability.fetch, resource=RedVideoResource)  # type: ignore
    async def fetch_resource(self, resource: RedResource) -> bytes:
        if (
            isinstance(resource, (RedImageResource, RedVoiceResource, RedVideoResource))
            and resource.path
            and resource.path.exists()
        ):
            with resource.path.open("rb") as f:
                return f.read()
        if isinstance(resource, RedImageResource):
            with suppress(Exception):
                async with ClientSession() as session:
                    async with session.get(resource.url) as resp:
                        return await resp.read()
        if TYPE_CHECKING:
            assert isinstance(resource.ctx.account, RedAccount)
        return await resource.ctx.account.websocket_client.call_http(
            "post",
            "api/message/fetchRichMedia",
            {
                "msgId": list(resource.ctx.cache["meta"].keys())[0].last_value,
                "chatType": 2 if resource.ctx.scene.last_key == "group" else 1,
                "peerUid": resource.ctx.scene.last_value,
                "elementId": resource.elem,
                "thumbSize": 0,
                "downloadType": 2,
            },
            raw=True,
        )
